;ͻ
;  CHANGE WORKING DIRECTORY By Christy Gemmell   Version 1.5 | June 1991    
;Ķ
;  CWD is a utility which enables you to jump from one directory to another 
;  repeatedly, without having to type in their pathnames every time. Use it 
;  as you would the normal CHDIR (Change Directory) command. The difference 
;  is that CWD stores the pathname of your current directory, in the MS-DOS 
;  environment area, before it takes you to the new one. Then when you want 
;  to return to the previous directory, enter 'CWD' without a pathname, and 
;                                 you're there!                             
;ͼ
;
;   Adapted and embellished from an idea by John B.Harrell III which was
;   published in 80-Micro Magazine, August 1987.
;
;   Ver 1.3 adds support for DOS 3.3 which moves the Master Environment a
;           further 64 bytes into the DOS environment block.
;
;   Ver 1.4 is rewritten to provide compatibility for DOS 4.0 which can have
;           additional Memory Control Blocks below the resident portion of
;           the command processor.
;
;   Ver 1.5 is rewritten to cater for DOS 5.0 which removes most of itself
;           into high memory on systems with 286 or better microprocessors.
;
;Ŀ
;          Constants and Equates.                                           
;
;
EnvVarLn    equ 07h
WriteString equ 09h                         ; Write String to Screen
DosCall     equ 21h                         ; DOS function processor
ExecInt     equ 2Eh                         ; Entry into COMMAND.COM
ChDir       equ 3Bh                         ; Move to a new Directory
GetDir      equ 47h                         ; Current directory pathname
SetBlock    equ 4Ah                         ; Set memory block
ShutDown    equ 4Ch                         ; Terminate a process
GetMCBPtr   equ 52h                         ; Get memory block pointer

;Ŀ
;          MASM MetaStatements.                                             
;
;
code        segment para 'code'
            assume  cs:code,ds:code,ss:code,es:nothing
            org     0080h

;Ŀ
;          Command Line.                                                    
;
;
PSP_Count   db   ?                          ; Command line character count
PSP         db   127 dup(?)                 ; Command line tail

;Ŀ
;          Program Code.                                                    
;
;
Chwd        proc  near
            mov   ah,30h                    ; DOS Service 48
            int   21h                       ;  - get version
            cmp   al,4                      ; Major version 4 or above?
            jae   Cwd_01                    ; Yes, okay to proceed
            jmp   Cwd_17                    ; Else report error
Cwd_01:
            mov   Major,al                  ; Save version for later  
            cld                             ; Clear direction flag
            mov   sp,offset StackArea       ; Switch to internal stack
            mov   bx,offset LastByte        ; Point to end of program
            shr   bx,1                      ; Shrink memory area
            shr   bx,1                      ;   reserved for this
            shr   bx,1                      ;     program down to
            shr   bx,1                      ;       minimal size
            inc   bx                        ; Allow for partial paragraphs
            mov   ah,SetBlock               ; DOS Service 74
            int   DosCall                   ;  - resize memory block
            jnc   Cwd_02                    ; Carry on if successful
            jmp   Cwd_13                    ; Fatal allocation error
Cwd_02:
            mov   si,offset DirBuffer       ; Point to storage area
            mov   dl,0                      ; Default drive
            mov   ah,GetDir                 ; DOS Service 71
            int   DosCall                   ;  - get current directory    
            jnc   Cwd_03                    ; Carry on if succesful
            jmp   Cwd_15                    ; Something's really wrong!
Cwd_03:
            mov   cx,64                     ; Length of directory buffer
            mov   di,si                     ; Set up string operation
            xor   al,al                     ; Set comparand
            repnz scasb                     ; Look for zero terminator
            mov   byte ptr [di-1],0Dh       ; Replace with C/R
            mov   bl,PSP_Count
            or    bl,bl                     ; Command tail?
            jz    Cwd_05                    ; No, return to last directory

;Ŀ
;          Set New Working Directory.                                       
;
;
            xor   bh,bh                     ; Index to command tail
            mov   byte ptr [PSP+bx],0       ; Terminate as ASCIIZ
            mov   di,offset PSP             ; Point to new pathname
            mov   cx,127                    ; Buffer count
            mov   al,' '                    ; Scan for leading blanks
            repz  scasb                     ;    leaving ES:DI pointing
            dec   di                        ;      to start of path
            cmp   byte ptr [di],0           ; Is anything there?
            jz    Cwd_05                    ; No, return to last directory  
            mov   dx,di                     ; DS:DX ==> ASCIIZ pathname
            mov   ah,ChDir                  ; Change current directory
            int   DosCall
            jnc   Cwd_04                    ; Jump if successful
            jmp   Cwd_14                    ; Report Error and exit
Cwd_04:
            call  Environment               ; Set LASTDIR
            mov   ah,ShutDown               ; Terminate a process
            xor   al,al                     ; No error
            int   DosCall                   ; Return to DOS

;Ŀ
;          Return to previously set Directory.                              
;
;
Cwd_05:
            mov   ah,GetMCBPtr              ; Get pointer to first memory
            int   DosCall                   ;   control block from DOS
            mov   ax,word ptr es:[bx-2]     ; AX = Master MCB segment
            mov   es,ax                     ; Check initial master block
            xor   bx,bx                     ;    for valid
            cmp   byte ptr es:[bx],4Dh      ;      header byte
            jnz   Cwd_16                    ; No, memory must be bad
Cwd_06:
            call  NextMCB                   ; Find next MCB
            mov   di,8                      ; ES:DI==> MCB owner name
            mov   cx,7                      ; Length of name field
            mov   si,offset Command         ; The name we're looking for
            repz  cmpsb                     ; Compare the two strings
            jnz   Cwd_06                    ; Not it, try the next one
            call  NextMCB                   ; Master environment is at
            cmp   Major,5                   ;    the next MCB under
            jb    Cwd_07                    ;      MS-DOS 4.x
            call  NextMCB                   ; Next but one for DOS 5.0
Cwd_07:
            mov   di,8                      ; ES:DI==> Environment table
Cwd_08:
            cmp   byte ptr es:[di],0        ; End of Environment string?
            jz    Cwd_12                    ; Yes, no LASTDIR set
            mov   cx,EnvVarLn               ; Set up for string comparison
            mov   si,offset EnvName         ; Point to test string
            repz  cmpsb                     ; Scan string
            jnz   Cwd_09                    ; Strings do not match
            cmp   byte ptr es:[di],'='      ; Name ends with '='?
            jz    Cwd_10                    ; Yes, we've found it!
Cwd_09:
            mov   cx,32767                  ; Set CX to a big number
            xor   al,al                     ; Look for the end
            repnz scasb                     ;    of this string then
            jmp   Cwd_08                    ;      process the next one
Cwd_10:
            inc   di                        ; Point to pathname
            push  ds                        ; Swap DS and ES so that
            push  es                        ;    DS:SI==> new variable
            pop   ds                        ;      and
            pop   es                        ;        ES:DI==> target string
            mov   si,di                     ; Set up source string
            mov   di,offset LastDir         ; Point to target buffer
Cwd_11:
            lodsb                           ; Get next byte
            stosb                           ; Copy it to buffer
            or    al,al                     ; Check for ASCIIZ terminator
            jnz   Cwd_11                    ; For length of string
Cwd_12:
            push  cs                        ; Restore DS
            pop   ds
            mov   dx,offset LastDir         ; DS:DX==>Last directory path
            mov   ah,ChDir                  ; Change current directory
            int   DosCall
            jc    Cwd_14                    ; Invalid, report Error
            call  Environment               ; Set new LASTDIR
            mov   ah,ShutDown               ; Terminate process
            mov   al,0                      ; Return without error
            int   DosCall

;Ŀ
;              Error Handling                                               
;
;
Cwd_13:
            mov   dx,offset Error1Msg       ; Memory allocation error
            jmp   short Cwd_18
Cwd_14:
            mov   dx,offset Error2Msg       ; Invalid Directory path
            jmp   short Cwd_18
Cwd_15:
            mov   dx,offset Error3Msg       ; Unanticipated Error
            jmp   short Cwd_18
Cwd_16:
            mov   dx,offset Error4Msg       ; Corrupted allocation block
            jmp   short Cwd_18
Cwd_17:
            mov   dx,offset Error5Msg       ; Wrong DOS version
Cwd_18:
            mov   ah,WriteString            ; Report error via DOS
            int   DosCall
            mov   ah,ShutDown               ; Terminate process
            mov   al,1                      ; Set Error Level
            int   DosCall                   ;    and exit
Chwd        endp

;Ŀ
;          Set current Directory path in DOS Environment                    
;
;
Environment proc  near
            mov   di,offset SetString+1     ; Find terminal <C/R>
            push  di
            mov   cx,32767                  ; Large value for CX
            mov   al,13                     ; Character to seek
            repnz scasb                     ; Go find it
            mov   bx,di
            pop   di
            sub   bx,di                     ; Calculate length
            dec   bx                        ;    of command string
            mov   byte ptr SetString,bl     ; Store length in command line
            mov   word ptr StackPtr,sp      ; Save stack pointer
            mov   word ptr StackSeg,ss      ;    and stack segment
            mov   si,offset SetString       ; Point to command string
            int   ExecInt                   ; Call primary COMMAND.COM
            mov   sp,word ptr cs:StackPtr   ; Restore stack pointer
            mov   ss,word ptr cs:StackSeg   ;    and segment register
            ret
Environment endp

;Ŀ
;          Find next Memory Control Block.                                  
;
;
NextMCB     proc  near
            mov   es,ax                     ; Block segment to ES
            xor   bx,bx                     ; ES:BX ==> Header byte
            inc   ax                        ; Increment starting segment
            mov   dx,word ptr es:[bx+3]     ; Get size of this block
            add   ax,dx                     ; Address of next block
            mov   es,ax                     ; Segment to ES
            cmp   byte ptr es:[bx],4Dh      ; Test for proper header
            jnz   Cwd_16                    ; Not a valid block, abort
            ret
NextMCB     endp

;Ŀ
;          Messages, Data and Working Stack.                                
;
;
Error1Msg   db    7,7,"CWD: I need a shrink!",13,10,"$"
Error2Msg   db    7,7,"CWD: You can't get there from here!",13,10,"$"
Error3Msg   db    7,7,"CWD: Where am I?",13,10,"$"
Error4Msg   db    7,7,"CWD: I've lost my memory!",13,10,"$"
Error5Msg   db    7,7,"CWD: This version is for DOS 4.0 or greater!"
            db    13,10,"$"

Major       db    ?                         ; DOS major version  

Command     db    'COMMAND'                 ; MCB name of COMMAND.COM

SetString   db    ' SET '                   ; String for EXEC
EnvName     db    'LASTDIR'
            db    '=\'
DirBuffer   db    64 dup(?)                 ; Buffer for old directory path
LastDir     db    '\',0                     ; Default to root directory
            db    62 dup(?)                 ; Buffer for environment path
StackSeg    dw    ?
StackPtr    dw    ?
            dw    128 dup(?)                ; Internal program stack
StackArea   equ   $
LastByte    equ   $                         ; Last byte of program + 1

code        ends
            end   Chwd

;Ŀ
;  Copyright 1988,1991 by Christy Gemmell and Singular Software.            
;
